@@ -9,17 +9,18 @@ import Control.Monad.Except
 import Control.Monad.Reader
 import Control.Monad.State.Strict
 import qualified Data.ByteString.Char8 as BS
+import qualified Data.Map as Map
 import Data.ByteString.Char8 (ByteString)
 import Data.CaseInsensitive (mk)
 import Data.Char (isAlphaNum, isAscii, isLetter, isSpace, ord, isPrint)
-import qualified Data.Map as Map
-import Data.Maybe (catMaybes, fromMaybe)
+import Data.Maybe (catMaybes)
 import Data.Monoid
 import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as Text
 import Data.Text.Encoding (decodeUtf8', encodeUtf8)
 import qualified Data.ByteString.UTF8 as UTF8
+import qualified Data.Attoparsec.ByteString.Char8 as A
 import Debug.Trace
 import Skylighting.Regex
 import Skylighting.Types
@@ -49,7 +50,8 @@ data TokenizerState = TokenizerState{
   , column              :: Int
   , lineContinuation    :: Bool
   , firstNonspaceColumn :: Maybe Int
-} deriving (Show)
+  , compiledRegexes     :: Map.Map RE Regex
+}
 
 -- | Configuration options for 'tokenize'.
 data TokenizerConfig = TokenizerConfig{
@@ -132,6 +134,7 @@ startingState =
                 , column = 0
                 , lineContinuation = False
                 , firstNonspaceColumn = Nothing
+                , compiledRegexes = Map.empty
                 }
 
 tokenizeLine :: (ByteString, Int) -> TokenizerM [Token]
@@ -212,15 +215,13 @@ tryRule rule inp = do
                 AnyChar cs -> withAttr attr $ anyChar cs inp
                 RangeDetect c d -> withAttr attr $ rangeDetect c d inp
                 RegExpr re -> withAttr attr $ regExpr (rDynamic rule) re inp
-                Int -> withAttr attr $ regExpr False integerRegex inp
-                HlCOct -> withAttr attr $ regExpr False octRegex inp
-                HlCHex -> withAttr attr $ regExpr False hexRegex inp
-                HlCStringChar -> withAttr attr $
-                                     regExpr False hlCStringCharRegex inp
-                HlCChar -> withAttr attr $ regExpr False hlCCharRegex inp
-                Float -> withAttr attr $ regExpr False floatRegex inp
-                Keyword kwattr kws ->
-                  withAttr attr $ keyword kwattr kws inp
+                Int -> withAttr attr $ parseInt inp
+                HlCOct -> withAttr attr $ parseOct inp
+                HlCHex -> withAttr attr $ parseHex inp
+                HlCStringChar -> withAttr attr $ parseCStringChar inp
+                HlCChar -> withAttr attr $ parseCChar inp
+                Float -> withAttr attr $ parseFloat inp
+                Keyword kwattr kws -> withAttr attr $ keyword kwattr kws inp
                 StringDetect s -> withAttr attr $
                                     stringDetect (rCaseSensitive rule) s inp
                 WordDetect s -> withAttr attr $
@@ -267,33 +268,6 @@ withAttr tt p = do
      then return Nothing
      else return $ Just (tt, res)
 
-hlCStringCharRegex :: RE
-hlCStringCharRegex = RE{
-    reString = reHlCStringChar
-  , reCompiled = Just $ compileRegex False reHlCStringChar
-  , reCaseSensitive = False
-  }
-
-reHlCStringChar :: ByteString
-reHlCStringChar = "\\\\(?:[abefnrtv\"'?\\\\]|[xX][a-fA-F0-9]+|0[0-7]+)"
-
-hlCCharRegex :: RE
-hlCCharRegex = RE{
-    reString = reStr
-  , reCompiled = Just $ compileRegex False reStr
-  , reCaseSensitive = False
-  }
-  where reStr = "'(?:" <> reHlCStringChar <> "|[^'\\\\])'"
-
-{-
--- Try a tokenizer and return to original state if it fails.
--- This is used when one tokenizer calls another.
-try :: TokenizerM a -> TokenizerM a
-try tokenizer' = do
-  origstate <- get
-  tokenizer' <|> (put origstate >> mzero)
--}
-
 wordDetect :: Bool -> Text -> ByteString -> TokenizerM Text
 wordDetect caseSensitive s inp = do
   wordBoundary inp
@@ -442,15 +416,22 @@ regExpr dynamic re inp = do
   reStr <- if dynamic
               then subDynamic (reString re)
               else return (reString re)
-  -- note, for dynamic regexes rCompiled == Nothing:
-  let regex = fromMaybe (compileRegex (reCaseSensitive re) reStr)
-                 $ reCompiled re
   when (BS.take 2 reStr == "\\b") $ wordBoundary inp
+  regex <- if dynamic
+              then return $ compileRegex (reCaseSensitive re) reStr
+              else do
+                compiledREs <- gets compiledRegexes
+                case Map.lookup re compiledREs of
+                     Nothing -> do
+                       let cre = compileRegex (reCaseSensitive re) reStr
+                       modify $ \st -> st{ compiledRegexes =
+                             Map.insert re cre (compiledRegexes st) }
+                       return cre
+                     Just cre -> return cre
   case matchRegex regex inp of
        Just (match:capts) -> do
-         match' <- decodeBS match
          modify $ \st -> st{ captures = capts }
-         takeChars (Text.length match')
+         takeChars (UTF8.length match)
        _ -> mzero
 
 wordBoundary :: ByteString -> TokenizerM ()
@@ -464,7 +445,11 @@ wordBoundary inp = do
 -- TODO is this right?
 isWordBoundary :: Char -> Char -> Bool
 isWordBoundary c d =
-  (isAlphaNum c && not (isAlphaNum d)) || (isAlphaNum d && not (isAlphaNum c))
+  (isAlphaNum c && not (isAlphaNum d))
+  || (isAlphaNum d && not (isAlphaNum c))
+  || (isSpace d && not (isSpace c))
+  || (isSpace c && not (isSpace d))
+
 
 decodeBS :: ByteString -> TokenizerM Text
 decodeBS bs = case decodeUtf8' bs of
@@ -523,10 +508,9 @@ keyword kwattr kws inp = do
   guard $ not (BS.null w)
   w' <- decodeBS w
   let numchars = Text.length w'
-  case kws of
-       CaseSensitiveWords ws   | w' `Set.member` ws -> takeChars numchars
-       CaseInsensitiveWords ws | mk w' `Set.member` ws -> takeChars numchars
-       _                       -> mzero
+  if w' `inWordSet` kws
+     then takeChars numchars
+     else mzero
 
 normalizeHighlighting :: [Token] -> [Token]
 normalizeHighlighting [] = []
@@ -536,35 +520,110 @@ normalizeHighlighting ((t,x):xs)
     (t, Text.concat (x : map snd matches)) : normalizeHighlighting rest
     where (matches, rest) = span (\(z,_) -> z == t) xs
 
-integerRegex :: RE
-integerRegex = RE{
-    reString = intReStr
-  , reCompiled = Just $ compileRegex False intReStr
-  , reCaseSensitive = False
-  }
-  where intReStr = "\\b[-+]?(0[Xx][0-9A-Fa-f]+|0[Oo][0-7]+|[0-9]+)\\b"
+parseCStringChar :: ByteString -> TokenizerM Text
+parseCStringChar inp = do
+  case A.parseOnly (A.match pCStringChar) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
-floatRegex :: RE
-floatRegex = RE{
-    reString = floatReStr
-  , reCompiled = Just $ compileRegex False floatReStr
-  , reCaseSensitive = False
-  }
-  where floatReStr = "\\b[-+]?(([0-9]+\\.[0-9]*|[0-9]*\\.[0-9]+)([Ee][-+]?[0-9]+)?|[0-9]+[Ee][-+]?[0-9]+)\\b"
+pCStringChar :: A.Parser ()
+pCStringChar = do
+  _ <- A.char '\\'
+  next <- A.anyChar
+  case next of
+       c | c == 'x' || c == 'X' -> () <$ A.takeWhile1 (A.inClass "0-9a-fA-F")
+         | c == '0' -> () <$ A.takeWhile1 (A.inClass "0-7")
+         | A.inClass "abefnrtv\"'?\\" c -> return ()
+         | otherwise -> mzero
 
-octRegex :: RE
-octRegex = RE{
-    reString = octRegexStr
-  , reCompiled = Just $ compileRegex False octRegexStr
-  , reCaseSensitive = False
-  }
-  where octRegexStr = "\\b[-+]?0[Oo][0-7]+\\b"
+parseCChar :: ByteString -> TokenizerM Text
+parseCChar inp = do
+  case A.parseOnly (A.match pCChar) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
-hexRegex :: RE
-hexRegex = RE{
-    reString = hexRegexStr
-  , reCompiled = Just $ compileRegex False hexRegexStr
-  , reCaseSensitive = False
-  }
-  where hexRegexStr = "\\b[-+]?0[Xx][0-9A-Fa-f]+\\b"
+pCChar :: A.Parser ()
+pCChar = do
+  () <$ A.char '\''
+  pCStringChar <|> () <$ A.satisfy (\c -> c /= '\'' && c /= '\\')
+  () <$ A.char '\''
+
+parseInt :: ByteString -> TokenizerM Text
+parseInt inp = do
+  wordBoundary inp
+  case A.parseOnly (A.match (pHex <|> pOct <|> pDec)) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r) -- assumes ascii
+
+pDec :: A.Parser ()
+pDec = do
+  mbMinus
+  _ <- A.takeWhile1 (A.inClass "0-9")
+  guardWordBoundary
+
+parseOct :: ByteString -> TokenizerM Text
+parseOct inp = do
+  wordBoundary inp
+  case A.parseOnly (A.match pHex) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r) -- assumes ascii
+
+pOct :: A.Parser ()
+pOct = do
+  mbMinus
+  _ <- A.char '0'
+  _ <- A.satisfy (A.inClass "Oo")
+  _ <- A.takeWhile1 (A.inClass "0-7")
+  guardWordBoundary
+
+parseHex :: ByteString -> TokenizerM Text
+parseHex inp = do
+  wordBoundary inp
+  case A.parseOnly (A.match pHex) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r) -- assumes ascii
+
+pHex :: A.Parser ()
+pHex = do
+  mbMinus
+  _ <- A.char '0'
+  _ <- A.satisfy (A.inClass "Xx")
+  _ <- A.takeWhile1 (A.inClass "0-9a-fA-F")
+  guardWordBoundary
+
+guardWordBoundary :: A.Parser ()
+guardWordBoundary = do
+  mbw <- A.peekChar
+  case mbw of
+       Just c ->  guard $ isWordBoundary '0' c
+       Nothing -> return ()
+
+mbMinus :: A.Parser ()
+mbMinus = (() <$ A.char '-') <|> return ()
+
+mbPlusMinus :: A.Parser ()
+mbPlusMinus = () <$ A.satisfy (A.inClass "+-") <|> return ()
+
+parseFloat :: ByteString -> TokenizerM Text
+parseFloat inp = do
+  wordBoundary inp
+  case A.parseOnly (A.match pFloat) inp of
+       Left _ -> mzero
+       Right (r,_) -> takeChars (BS.length r)  -- assumes all ascii
+  where pFloat :: A.Parser ()
+        pFloat = do
+          let digits = A.takeWhile1 (A.inClass "0-9")
+          mbPlusMinus
+          before <- A.option False $ True <$ digits
+          dot <- A.option False $ True <$ A.satisfy (A.inClass ".")
+          after <- A.option False $ True <$ digits
+          e <- A.option False $ True <$ (A.satisfy (A.inClass "Ee") >>
+                                         mbPlusMinus >> digits)
+          mbnext <- A.peekChar
+          case mbnext of
+               Nothing   -> return ()
+               Just c    -> guard (not $ A.inClass "." c)
+          guard $ (before && not dot && e)     -- 5e2
+               || (before && dot && (after || not e)) -- 5.2e2 or 5.2 or 5.
+               || (not before && dot && after) -- .23 or .23e2
 
